#include <complex>

#include "prconst.h"
#include "Air.h"
#include "water.h"

/** Default Ctor.
* @param[in] temp
* @param[in] press
*/
Air::Air(const double temp, const double press) : T(temp), Panom(press)
{
	mix=NULL ;
	mixcnt=0 ;
}

/** Constructor with files defining constituents.
* @param[in] temp
* @param[in] press
* @param[in] relhum
* @param[in] co2
* @param[in] h2o
* @param[in] useV
* @param[in] dir
* @param[in] fcount
* @param[in] 
*/
Air::Air(const double temp, const double press, const double relhum, const double co2, const double o2, const double h2o,
	const double o3, const bool useV, const string dir, int fcount, char *argv[]) : T(temp), Panom(press), mixcnt(fcount)
{
	mix=new compon[mixcnt] ;
	double mols ;
	for(int com=0; com<mixcnt; com++)		// each argument denotes one molecule type
	{
		char *fname=argv[com] ;
		mix[com].molc = new molecule(fname,dir) ;
		const int molec= mix[com].molc->molec ;

		/* the strategy: first set compon::volpercent, then further down compon::mol_m3
		*/
		switch (molec)
		{
		case 2:
			mix[com].volpercent = co2 ;
			break ;
		case 3:
			mix[com].volpercent = o3 ;
			break ;
		case 7:
			mix[com].volpercent = o2 ;
			break ;
		default:
			mix[com].volpercent = volrel[molec-1] ;
		}
		if ( molec == 1 && h2o > 0. )
			mix[com].volpercent = h2o ;

		/* Decision on to use or not to use the virials.
		*/
		virial * virials ;
		if ( useV) 
			virials = new virial(molec) ;
		else
			virials = new virial() ;

		switch(molec)
		{
		case 1:		// water; molecules per cubic meter
			{
				water * H2o  ;
				if ( useV) 
					H2o = new water(T,WPruss) ;
				else
					H2o = new water(T,*virials) ;
				if ( h2o < 0.)
					mix[com].mol_m3 = AVOGADRO*H2o->n(relhum) ;
				else
					mix[com].mol_m3 = mix[com].volpercent*AVOGADRO*virials->p2n(Panom,T) ;
				delete H2o ;
			}
			break ;
		case 3:		
				mix[com].mol_m3 = mix[com].volpercent*AVOGADRO*virials->p2n(Panom,T) ;
			break ;
		default :
			mix[com].mol_m3 = mix[com].volpercent*AVOGADRO*virials->p2n(Panom,T) ;
		}

		/* ideal gase equation is p=(n/V)*R*T, in contrast.
		*/
		mix[com].partp = PA2MMHG( virials->n2p(mix[com].mol_m3/AVOGADRO,T) ) ;

		delete virials ;
	}

	if( scale_p)
	{
		/* use the sign of the value of h2o to choose either between scaling all dry components,
		* or scaling all components besides CO2, O2, and H2O
		*/
		if ( h2o >= 0.)
		{
			set<int> excl;
			excl.insert(1) ; excl.insert(2) ; excl.insert(7) ;
			scale(useV,excl) ;
		}
		else
			scaledry(useV) ;
	}
	
	for(int com=0; com<mixcnt; com++)		// each argument denotes one molecule type
	{
		const int molec= mix[com].molc->molec ;
		cout << "# " << nam[molec-1] << " (" << molec << "), " << mix[com].molc->transcnt << " lines, " 
			<< mix[com].mol_m3 << " molec/m^3," << mix[com].mol_m3/AVOGADRO << " mole/m^3," ;
		cout << " " << MMHG2PA(mix[com].partp)/100. << " hPa\n" ;
	}
	cout << "# total " << Pa(mols,false) << " Pa, " ;
	cout << mols << " mols/m^3" << endl ;
}

Air::~Air()
{
	//for( int indx=0 ; indx < mixcnt ; indx++)
	//	delete [] mix[indx].molc ;
	delete [] mix ;
}

#ifdef USE_QOFT
void Air::Qsum()
{
	for(int com=0 ; com < mixcnt;com++)
		mix[com].molc->Qsum(T) ;
}
#endif

// Lorentz line profile, returns susceptibility
complex<double> Air::suscep(const double lambda)
{
	complex<double> resul ;
	for(int com=0 ; com < mixcnt;com++)
		resul += mix[com].mol_m3*mix[com].molc->Lorentz(lambda) ;
	resul *= ECHARGE*ECHARGE/(EMASS*EPSILONZERO) ;
	return resul /(1.-resul/3.) ;
}



/** Compute the magnetic susceptibility.
* @param[in] useV if true, use virials in the equation of state
* @see
\latexonly
\cite{DavisMet35}
\endlatexonly
*/
double Air::magn(const double useV)
{
	/* Table 3 last column
	*/
	const double alpha = 3.1251e-7 ;
	const double beta = 4.870 ;
	const double gamma = 4.607e-5 ;
	const double delta = 1.018 ;

	const int h2oindx = hash2o(1) ;	// if non-negative, this may be not unique if we have accumulated sources
	double mols ;
	const double p = Pa(mols,false) ;
	double xv = 0. ;
	if( h2oindx != -1)
	{
		virial *virs ;
		if ( useV)
			virs = new virial(1) ;
		else
			virs = new virial() ;
		// we aproximate the mole fraction by the fraction of the total pressure
		xv = virs->n2p(mix[h2oindx].mol_m3/AVOGADRO,T) / p ;
		delete virs ;
	}
	return alpha*p/(T*T)*(1.-delta*xv)*(1.-beta*(-0.0004)-gamma*(KELVIN2CELS(T)-20.)) ;
}

/** Return the volume fraction (roughly 370) of carbon dioxide
* if it was in the list of molecules included. Otherwise return zero.
*@return CO2 volume fraction in ppmv
*/
double Air::ppmCO2() const
{
	for(int com=0 ; com < mixcnt; com++)
	{
		if (mix[com].molc->molec == 2 )
			/* The variable \c volpercent is in units of parts per one. Multiply
			* with 10^6 to return ppmv
			*/
			return 1.e6*mix[com].volpercent ;
	}
	return 0. ;
}


/** Compute the total (external) pressure.
*@param[in] excludh2o if true, any contribution of H2O is ignored (skipped, implicitly zero)
*@return pressure [Pa]
*/
double Air::Pa(double &mols, const bool excludh2o) const
{
	double Pasum=0. ;
	mols =0. ;
	set<int> typs ;
	for(int com=0 ; com < mixcnt;com++)
	{
		const int molec= mix[com].molc->molec ;
		if( typs.find(molec) == typs.end() && (excludh2o == false || molec != 1) )
		{
			typs.insert(molec) ;
			Pasum += MMHG2PA(mix[com].partp) ;
			mols +=  mix[com].mol_m3/AVOGADRO ;
		}
	}
	return Pasum ;
}

/** Check if water is in the molecular list.
* @param[in] molecNo the HITRAN molecular number to search for.
*     Chose 1 for water, for example.
* @return the index (position) of water in the line list. -1 if absent.
*/
int Air::hash2o(const int molecNo) const
{
	for(int com=0 ; com < mixcnt;com++)
		if (mix[com].molc->molec == molecNo )
			return com ;
	return -1 ;
}


/**
* @param[in] useV if true, use virials in the equation of state, else ideal gas law
*/
void Air::scaledry(const bool useV)
{
		// check whether we actually accumulated the nominal pressure; if not, scale the dry components to achieve
	double presscal = 1. ;
	const int h2oindx = hash2o(1) ;	// if non-negative, this may be not unique if we have accumulated sources
	double mols ;

	/* loop 2 times for self-consistent precision, since this is a non-linear
	* problem if virials are in use.
	*/
	for( int scloop = 0 ; scloop < 2 ; scloop++)
	{
		if ( h2oindx != -1)
		{
			virial *virs ;
			if ( useV)
				virs = new virial(1) ;
			else
				virs = new virial() ;
			presscal = (Panom- virs->n2p(mix[h2oindx].mol_m3/AVOGADRO,T))/Pa(mols,true) ;
			delete virs ;
		}
		else
			presscal = Panom/Pa(mols,true) ;

		for(int com=0 ; com < mixcnt;com++)
		{
			const int molec= mix[com].molc->molec ;
			if( molec != 1)
			{
				mix[com].mol_m3 *= presscal ;
				mix[com].volpercent *= presscal ;
			}
			virial *virs ;
			if ( useV)
				virs = new virial(molec) ;
			else
				virs = new virial() ;
			mix[com].partp = PA2MMHG( virs->n2p(mix[com].mol_m3/AVOGADRO,T) ) ;	// p=(n/V)*R*T
			delete virs ;
		}
	}
}

/**
* @param[in] useV if true, use virials in the equation of state, else ideal gas law
* @param[in] excl list of HITRAN molecular numbers excluded from re-scaling
* @since 2008-03-19
*/
void Air::scale(const bool useV, const set<int> excl)
{

	/* loop 2 times for self-consistent precision, since this is a non-linear
	* problem if virials are in use.
	*/
	for( int scloop = 0 ; scloop < 2 ; scloop++)
	{
		/* Calculate the sum of the partial pressures of the molecules which are not
		* excluded from scaling in Pasum, the sum of the others in Paexcl.
		* Take precaution not to double count molecules used with more than one line list.
		*/
		double Pasum = 0. ;
		double Paexcl = 0. ;
		set<int> typs ;
		for(int com=0 ; com < mixcnt;com++)
		{
			const int molec= mix[com].molc->molec ;
			if( typs.find(molec) == typs.end() )
			{
				typs.insert(molec) ;
				if ( excl.find(molec) == excl.end() )
					Pasum += MMHG2PA(mix[com].partp) ;
				else
					Paexcl += MMHG2PA(mix[com].partp) ;
			}
		}

		/* the aim is to have Pasum+Paexcl=Panom. If this is not the case,
		* set presscal to end up with Pasum*presscal+Paexcl=Panom.
		*/
		const double presscal = (Panom-Paexcl)/Pasum ;

		for(int com=0 ; com < mixcnt;com++)
		{
			const int molec= mix[com].molc->molec ;
			if( excl.find(molec) == excl.end() )
			{
				mix[com].mol_m3 *= presscal ;
				mix[com].volpercent *= presscal ;
			}
			virial *virs ;
			if ( useV)
				virs = new virial(molec) ;
			else
				virs = new virial() ;
			mix[com].partp = PA2MMHG( virs->n2p(mix[com].mol_m3/AVOGADRO,T) ) ;	// p=(n/V)*R*T
			delete virs ;
		}
	}
}
bool Air::scale_p ;


const double Air::volrel[] = {
	0.,3.7e-4,1.e-7,3.0e-7,1.0e-7,1.6e-6,0.20983,5.e-10,5.e-10,1.e-9,
	4.e-7, 0., 3.9e-13, 0.,0.,0.,0.,0.,0.,0.,
	0.,0.781, 0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,1.8e-5,
	1.e-6,9.34e-3,5.e-7,5.2e-6
	} ;

const char * Air::nam[] = {
		"H2O","CO2","O3","N2O","CO","CH4","O2","NO","SO2","NO2",  /* 0-9 */
		"NH3","HNO3","OH","HF","HCl","HBr","HI","ClO","OCS","H2CO",       	/* 10-19 */
		"HOCl","N2","HCN","CH3Cl","H2O2","C2H2","C2H6","PH3","COF2","SF6",		/* 20-29 */
		"H2S","HCOOH","HO2","O","ClONO2","NO+","HOBr","C2H4","CH3OH","",		/* 30-39 */
		"","","","","","","","","","",		/* 40-49 */
		"","","","","","","","","","",		/* 50-59 */
		"","","","","","","","","","",		/* 60-69 */
		"","","","","","","","","","",		/* 70-79 */
		"","","","","","","","","","",		/* 80-89 */
		"","","","","","","","","","Ne",	/* 90-99 */
		"Kr","Ar","H2","He"} ;			/* 100-109 */

