#include <cmath>
#include "prconst.h"
#include "WagnerPruss.h"

using namespace std ;

/** Ctor.
* @param[in] kelvin temperature of water (Kelvin)
*/
WagnerPruss::WagnerPruss(double kelvin) : T(kelvin)
{
	tau = Tc/T ;
	pSatur = pWPruss() ;
}

/** Calculate water density.
* @param[in] humidity relative humidity in percent
* @return water density in moles per cubic meter
*/
double WagnerPruss::n(const double humidity) const
{
	double pascal ;	// partial pressure
	return rho(humidity)/M ;
}

/** Calculate partial pressure.
* @param[in] humidity humidity between 0 and 100 (percent)
* @return partial pressure [Pa]
*/
double WagnerPruss::P(const double humidity) const
{
	return 0.01*humidity*pSatur ;
}

/** Calculate density.
* @param[in] humidity
* @return density in gram per cubic meter
*/
double WagnerPruss::rho(const double humidity) const
{
	// partial pressure obtained from the saturated vapor curve at the current temperature
	const double p = P(humidity) ;
	// first equation table 6.3 page 431:  rho = p/[R*T*(1+delta Phi_delta^r)],
	// Since delta =rho/rhoc, this is a self-consistent equation...
	// p is in Pa, R in J/g/K and rhoc in g/m^3
	double delta=p/(R*T*rhoc) ;
	double Phideriv = Phi_delta_r(delta) ;
	for(int loo=1; loo<3 ; loo++)
	{
		delta = p/(R*T*rhoc*(1.+delta*Phideriv)) ;
		Phideriv = Phi_delta_r(delta) ;
		// cerr << __FILE__ << " " << __LINE__ << " " << loo << " " << delta << " " << Phideriv << endl ;
	}
	/* Since R above was in unitsof kg/kJ/K, the
	* value returned is in gram per volume
	*/
	return p/(R*T*(1.+delta*Phideriv)) ;
}

/** Saturation pressure.
* @return saturation pressure [Pa]
* @see eq 2.5
\latexonly
\cite[p 398]{WagnerJPCR31}
\endlatexonly
*/
double WagnerPruss::pWPruss() const
{
	static double a[]={-7.85951783 ,1.84408259 , -11.7866497, 22.6807411, -15.9618719, 1.80122502} ;
	static double pc = 22.064e6 ;	// Pascal
	const double theta = 1.0-T/Tc ;
	return pc*exp(tau
	*(a[0]*theta+a[1]*pow(theta,1.5)+a[2]*pow(theta,3.)+a[3]*pow(theta,3.5)+a[4]*pow(theta,4.)+a[5]*pow(theta,7.5))) ;
}

/** Derivative of a thermodynamic potential.
* @return \f$\partial \Delta/\partial \delta\f$, table 6.5, page 434
\latexonly
\cite[p 398]{WagnerJPCR31}
\endlatexonly
*/
double WagnerPruss::dDeltaddelta(const int i, const double dmin1, const double theta) const
{
	const double dmin12 = dmin1*dmin1 ;	// (delta-1)^2
	double tmp = tab62c[i-55].Ai*theta/tab62c[i-55].beti *pow(dmin12,0.5/tab62c[i-55].beti-1.) ;
	tmp += tab62c[i-55].Bi*tab62c[i-55].ai *pow(dmin12,tab62c[i-55].ai-1.) ;
	return 2.*dmin1*tmp ;
}

/** Derivative of a thermodynamic potential.
* @return \f$\partial \Phi^r/\partial \delta\f$,
\latexonly
\cite[p 398]{WagnerJPCR31}
\endlatexonly
*/
double WagnerPruss::Phi_delta_r(const double delta) const
{
	double phideltar =0 ;
	for(int i=1; i<= 7; i++)
	{
		const double tmp = tab62a[i-1].ni * tab62a[i-1].di * pow(tau,tab62a[i-1].ti) ;
		if ( tab62a[i-1].di == 1)
			phideltar += tmp ;
		else
			phideltar += tmp*pow(delta,tab62a[i-1].di-1) ;
	}
	for(int i=8; i<= 51; i++)
	{
		const double dci= pow(delta,tab62a[i-1].ci) ;
		const double tmp= tab62a[i-1].ni*exp(-dci)*pow(tau,tab62a[i-1].ti)*(tab62a[i-1].di-tab62a[i-1].ci*dci) ;
		if ( tab62a[i-1].di == 1)
			phideltar += tmp ;
		else
			phideltar += tmp*pow(delta,tab62a[i-1].di-1) ;
	}
	// no contribution from there, because all di==3
	for(int i=52; i<= 54; i++)
	{
		const double depsi= delta-tab62b[i-52].epsi ;
		const double expo = tab62b[i-52].alphi*pow(depsi,2)+tab62b[i-52].beti*pow(tau-tab62b[i-52].gami,2) ;
		const double tmp= tab62b[i-52].ni*pow(delta,tab62b[i-52].di)*exp(-expo)
					*(tab62b[i-52].di/delta-2.*tab62b[i-52].alphi*depsi) ;
		if ( tab62b[i-52].ti == 0)
			phideltar += tmp ;
		else
			phideltar += tmp*pow(tau,tab62b[i-52].ti) ;
	}
	for(int i=55; i<= 56; i++)
	{
		// eq 6.6 ...
		const double dmin1 = delta-1. ;
		const double psi = exp(-tab62c[i-55].Ci*dmin1*dmin1-tab62c[i-55].Di*(tau-1.)*(tau-1.)) ;
		const double thet = 1.-tau+ tab62c[i-55].Ai *pow(dmin1*dmin1,0.5/tab62c[i-55].beti) ;	// after 6.6
		const double Delta = thet*thet + tab62c[i-55].Bi*pow(dmin1*dmin1,tab62c[i-55].ai) ;
		const double dPsidd = -2.*tab62c[i-55].Ci*dmin1*psi ; // bottom p 433, d(psi)/ddelta
		const double dDbidd = tab62c[i-55].bi*pow(Delta,tab62c[i-55].bi-1.)*dDeltaddelta(i,dmin1,thet) ;
		phideltar += tab62c[i-55].ni*
				( pow(Delta,tab62c[i-55].bi)*(psi+delta*dPsidd)+dDbidd*delta*psi ) ;
	}
	return phideltar ;
}


const WagnerPruss::tab62aS WagnerPruss::tab62a[] = {
	{ 0, 1, -0.5, 0.12533547935523e-1},
	{ 0, 1, 0.875, 0.78957634722828e1},
	{ 0, 1, 1., -0.87803203303561e1},
	{ 0, 2, 0.5, 0.31802509345418},
	{ 0, 2, 0.75, -0.26145533859358},
	{ 0, 3, 0.375, -0.78199751687981e-2},
	{ 0, 4, 1., 0.88089493102134e-2},
	{ 1, 1, 4., -0.66856572307965},
	{ 1, 1, 6., 0.20433810950965},
	{ 1, 1, 12., -0.66212605039687e-4},
	{ 1, 2, 1., -0.19232721156002},
	{ 1, 2, 5., -0.25709043003438},
	{ 1, 3, 4., 0.16074868486251},
	{ 1, 4, 2., -0.40092828925807e-1},
	{ 1, 4, 13., 0.39343422603254e-6},
	{ 1, 5, 9., -0.75941377088144e-5},
	{ 1, 7, 3., 0.56250979351888e-3},
	{ 1, 9, 4., 0.156086522571353e-4},
	{ 1, 10, 11., 0.11537996422951e-8},
	{ 1, 11, 4., 0.36582165144204e-6},
	{ 1, 13, 13., -0.13251180074668e-11},
	{ 1, 15, 1., -0.62639586912454e-9},
	{ 2, 1, 7., -0.10793600908932},
	{ 2, 2, 1., 0.17611491008752e-1},
	{ 2, 2, 9., 0.22132295167546},
	{ 2, 2, 10., -0.40247669763528},
	{ 2, 3, 10., 0.58083399985759},
	{ 2, 4, 3., 0.49969146990806e-2},
	{ 2, 4, 7., -0.31358700712549e-1},
	{ 2, 4, 10., -0.74315929710341},
	{ 2, 5, 10., 0.47807329915480},
	{ 2, 6, 6., 0.20527940895948e-1},
	{ 2, 6, 10., -0.13636435110343},
	{ 2, 7, 10., 0.14180634400617e-1},
	{ 2, 9, 1., 0.83326504880713e-2},
	{ 2, 9, 2., -0.29052336009585e-1},
	{ 2, 9, 3., 0.38615085574206e-1},
	{ 2, 9, 4., -0.20393486513704e-1},
	{ 2, 9, 8., -0.16554050063734e-2},
	{ 2, 10, 6., 0.19955571979541e-2},
	{ 2, 10, 9., 0.15870308324157e-3},
	{ 2, 12, 8., -0.16388568342530e-4},
	{ 3, 3, 16., 0.43613615723811e-1},
	{ 3, 4, 22., 0.34994005463765e-1},
	{ 3, 4, 23., -0.76788197844621e-1},
	{ 3, 5, 23., 0.22446277332006e-1},
	{ 4, 14, 10., -0.62689710414685e-4},
	{ 6, 3, 50., -0.55711118565645e-9},
	{ 6, 6, 44., -0.19905718354408},
	{ 6, 6, 46., 0.31777497330738},
	{ 6, 6, 50., -0.11841182425981} } ;
const WagnerPruss::tab62bS WagnerPruss::tab62b[] = { 
	{ 3, 0, -0.31306260323435e2, 20, 150, 1.21 ,1 },
	{ 3, 1, 0.31546140237781e2, 20, 150, 1.21 ,1 },
	{ 3, 4, -0.25213154341695e4, 20, 250, 1.25 ,1 } } ;
const WagnerPruss::tab62cS WagnerPruss::tab62c[] = { 
	{ 3.5, 0.85, 0.2, -0.14874640856724, 28, 700, 0.32 , 0.3 },
	{ 3.5, 0.95, 0.2, 0.31806110878444, 32, 800, 0.32 , 0.3 } } ;
