using namespace std ;

#include <complex>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstring>
#include <cstdlib>
#include <climits>
#include <unistd.h>

#include "prconst.h"
#include "molecule.h"

/** convert wavenumber to micron
*/
#define WN2MICRON(w) (1.e4/(w))

/** convert micron to angular frequency
*/
#define MICRON2OMEGA(m) (2.*M_PI*1.e6*VACUUMC/(m))

/** Directory of the HITRAN2004 data base files. If not an absolute
* path name, it is interpreted relativ to the executable.
* This contains files named 02_hit04.par, 04_hit06.par and so on.
* @see <a href="http://www.hitran.com">www.hitran.com</a> .
*/
#define HITRAN2000DIR "HITRAN/HITRAN2004/By-Molecule/"

/** Directory of the JPL data base.
* This contains files named c018003.cat and so on.
* @see <a href="http://spec.jpl.nasa.gov/ftp/pub/catalog">spec.jpl.nasa.gov</a> .
*/
#define SPECJPLDIR "spec/catalog/"

/** Directory of the Cfa data base.
* This contains files named arfvals.txt, c2ofvals.txt and so on.
* @see <a href="http://www.cfa.harvard.edu/amp/tools.html">www.cfa.harvard.edu</a> .
*/
#define CFAWWWDIR "cfa/"

/** Directory of the other data bases retrieved from other pieces of literature
*/
#define JCP111DIR "hitran_rjm/"

/** Directory of the GEISA data base.
* This contains files named geisa2003_01.gs and so on.
* The newer GEISA-09 format shows files with names like line_GEISA2009_asc_gs08...
* @see <a href="http://ara.lmd.polytechnique.fr/htdocs-public/products/GEISA/HTML-GEISA/GeisaAccess.html">ara.lmd.polytechnique.fr</a> .
*/
#define GEISADIR "GEISA/"

/** Maximum number of bytes per GEISA 2009 input line (Fortran total 252), which is followed by \n or \r\n, then \0
* In case of providing space for input data, this is to be considered larger than HITRANLL
*/
#define GEISALL 255

// default constructor: only the molecule number given
molecule::molecule(int molno) : molec(molno)
{
	isot=1 ;		// most abundant isotope by default
	trans=NULL ;
	transcnt=0 ;
	murange[0]=1.e6 ; murange[1]=0. ;
}

// constructor: read file with given name, select only lines with defined molecular number 'molno'
// If 'molno' isn't specified, let the file implicitly define the molecule as the one given in its first line
// iso=1 (the most abundant one) by default.
molecule::molecule(const char *fname, const string dir, int molno, int iso) : molec(molno), isot(iso)
{
	trans=NULL ;
	transcnt=0 ;
	int lineno=0 ;
	char fullfname[PATH_MAX+1] ;			// full UNIX path file name
	catalog cattype;
	if ( fname[0] >= '0' && fname[0] <= '9')	// pretach the base directory if an "original" file name
	{
							// was specified, the name of which starts with a number
		sprintf(fullfname,"%s/%s%s",dir.c_str(),HITRAN2000DIR,fname) ;
		cattype=HITRAN ;
	}
	else if ( strstr(fname,".cat") )		// names on ".cat" are spec.jpl.nasa.gov files
	{
		sprintf(fullfname,"%s/%s%s",dir.c_str(),SPECJPLDIR,fname) ;
		cattype=SPECJPL ;
	}
	else if ( fname[0] == 'J' )					// JCP names are JCP 111 no 6 p 2448 files
	{
		sprintf(fullfname,"%s/%s%s",dir.c_str(),JCP111DIR,fname) ;
		cattype=JCP111 ;
	}
	else if ( strstr(fname,".txt") )		// names on ".txt" are cfa-www.harvard.edu/amdata/ampdata files
	{
		sprintf(fullfname,"%s/%s%s",dir.c_str(),CFAWWWDIR,fname) ;
		cattype=CFAWWW ;
	}
	else
	{
		strcpy(fullfname,fname) ;
		cattype=HITRAN ;
	}
#ifdef __GNUC__
	ifstream hitrfile(fullfname) ;	//open DB file for reading
	if( ! hitrfile )
	{
		cerr << "Could not open " << fullfname << endl ;
		exit(EXIT_FAILURE) ;
	}
#else
	ifstream hitrfile(fullfname,ios::nocreate ) ;	//open DB file for reading
#endif
	// prescan: count number of relevant lines first, allocate space, then rewind file and scan again
	while( !hitrfile.eof() && hitrfile)		// read line by line
	{
		char linebuf[GEISALL] ;
		hitrfile.getline(linebuf,GEISALL) ;		// 100 characters plus \n in the files, maybe \r\n
		if( hitrfile.eof() )
			break ;
		if( linebuf[0] == '#')		// skip comment lines
			continue ;
		istringstream firstline(linebuf) ;
		transition thistrans(firstline,cattype) ;	// define transition in current input line
		if( lineno==0 && molec ==0)		// if in first input line and yet unspecified molecule number
			molec= thistrans.molec ;

		if( thistrans.molec == molec && thistrans.isot == isot ) // select only those with matching isotope specs
		{
			// update the range of selected microns...
			if( thistrans.micron < murange[0])
				murange[0]= thistrans.micron ;
			if( thistrans.micron > murange[1])
				murange[1]= thistrans.micron ;
			transcnt++ ;
		}
		lineno++ ;
	}
	hitrfile.clear() ;			// clear ios state (note: to be done before seekg()!
	hitrfile.seekg(0,ios::beg) ;			// rewind file
	trans= new transition[transcnt] ;		// allocate all in once
	lineno=transcnt=0 ;
	while( !hitrfile.eof() && hitrfile)		// read line by line
	{
		char linebuf[GEISALL] ;
		hitrfile.getline(linebuf,GEISALL) ;		// 100 characters plus \n in the files, maybe \r\n
		if( hitrfile.eof() )
			break ;
		if( linebuf[0] == '#')		// skip comment lines
			continue ;
		istringstream firstline(linebuf) ;
		transition thistrans(firstline,cattype) ;	// define transition in current input line
		if( lineno==0 && molec ==0)		// if in first input line and yet unspecified molecule number
			molec= thistrans.molec ;

		if( thistrans.molec == molec && thistrans.isot == isot ) // select only those with matching isotope specs
		{
			trans[transcnt] = thistrans ;	// transition constructor by stream
			transcnt++ ;
		}
		lineno++ ;
#ifdef DEBUG
		if ( lineno % 200 == 0 )
			cerr << lineno << " lines read\n" ;
#endif
	}
	// cout << "file " << fullfname << " with " << lineno << " lines, " << transcnt << " taken\n" ;
	molmass=mmas[molec-1] ;
}

// output into a stream in gnuplot style
// 'lowlim and 'hilim' define a filter region (microns) of values to include.
void molecule::gnuplot(const double lowlim, const double hilim , ostream & out)
{
	out << "# " << lowlim << " - " << hilim << " microns\n" ;
	out << "# molecule " << molec  << " isotope " << isot << "transcnt " << transcnt << endl ;
	for(int li=0 ;li < transcnt; li++)
		if( trans[li].micron >= lowlim && trans[li].micron <= hilim)
		{
			// wavenumber and micron are aligned on a 'reciprocal' scale: start with higher wavenum, then lower
			out << WN2MICRON(trans[li].waven+trans[li].width) << " " << trans[li].waven+trans[li].width << " 0.\n";
			out << trans[li].micron << " " << trans[li].waven << " " << trans[li].strength << "\n" ;
			out << WN2MICRON(trans[li].waven-trans[li].width) << " " << trans[li].waven-trans[li].width << " 0.\n\n";
		}
}

// Lorentz line profile, returns sum over all lines of f/[omega0^2-omega^2 -i Gamma Omega] where Gamma is FWHM
complex<double> molecule::Lorentz(const double lambda) const
{
	const double omega=MICRON2OMEGA(lambda) ;
	complex<double> resul(0.,0.) ;
	for(int li=0 ; li < transcnt;li++)
		resul += trans[li].Lorentz(omega) ;
	return resul ;
}



#ifdef USE_QOFT
void molecule::Qsum(double T)
{
	for(int li=0 ; li < transcnt;li++)
		if ( trans[li].lowsta > 0.)
			trans[li].Qsum(T) ;
}
#endif

molecule::~molecule()
{
	delete [] trans ;
}

/*
* molar masses, 1=H2O, 2=CO2, 3=O3, 4=N2O, 5=CO, 6=CH4, 7=O2, 8=NO, 9=SO2, 10=NO2, 11=NH3, 13=OH, 22=N2,
* ATOMS_OFFSET=Ne, ATOMS_OFFSET+1=Kr, ATOMS_OFFSET+2=Ar, ATOMS_OFFSET+3=H2, ATOMS_OFFSET+4=He
* see HITRAN's file MOLPARAM.TXT
*/
const double molecule::mmas[] = {
	18.010565, 43.989830, 47.984745, 44.001062, 27.994915, 16.031300, 31.989830, 29.997989, 63.961901, 45.992904,
	17.026549, 62.995644, 17.002740, 20.006229, 35.976678, 79.926160, 127.912297, 50.963768, 59.966986, 30.010565,
	51.971593, 28.006147, 27.010899, 49.992328, 0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
	0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,20.183 ,
	83.80   ,  39.948 ,   2.01565 ,  4.006
	} ;
